from mcconv import GenericTextReader, UnparsedTextEvent

import os
import unittest
from mcconv import detect_mc_type, McFileTypes



class TestGenericTextReader(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def _mc_file_path(self, path):
        """Gets data file path"""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dir_path, 'data', path)


    def test_read_file(self):
        """Test detecting BeaGLE file type"""

        def is_line_relevant(line):
            is_comment = any(word in line for word in ['BEAGLE', 'I', '='])
            return not is_comment

        def is_event(tokens):
            return len(tokens) > 20 # Beagle event header has like 50 tokens, particle like 16

        def is_particle(tokens):
            return len(tokens) < 20

        reader = GenericTextReader(is_line_relevant, is_event, is_particle)
        reader.open(self._mc_file_path('beagle_3events.txt'))

        events = [event for event in reader.events()]
        self.assertEqual(len(events), 3)
        self.assertEqual(events[0].start_line_index, 6)        # The first event is on line index 6
        self.assertEqual(len(events[0].unparsed_particles), 11)   # There are 11 particles
        self.assertEqual(events[0].unparsed_particles[0][2], "11")  # There are 11 particles
        self.assertEqual(events[0].event_tokens[6], "-5")  # There are 11 particles
        reader.close()

    def test_default_event_identification(self):
        """Test detecting BeaGLE file type"""

        reader = GenericTextReader()
        reader.open(self._mc_file_path('beagle_3events.txt'))

        events = [event for event in reader.events()]
        self.assertEqual(len(events), 3)
        self.assertEqual(events[0].start_line_index, 6)        # The first event is on line index 6
        self.assertEqual(len(events[0].unparsed_particles), 11)   # There are 11 particles
        self.assertEqual(events[0].unparsed_particles[0][2], "11")  # There are 11 particles
        self.assertEqual(events[0].event_tokens[6], "-5")  # There are 11 particles
        reader.close()

    def test_event_ranges(self):
        """Test number of events to skip and to take"""
        reader = GenericTextReader()
        reader.open(self._mc_file_path('pythia6-radcor-10evt.txt'))

        events = [event for event in reader.events(evt_skip=3, evt_take=2)]
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].start_line_index, 167)  # The 4th event is on line index 168
        self.assertEqual(events[0].unparsed_particles[2][3], '-211')  #



from .file_types import McFileTypes
import pathlib


def detect_mc_type(file_name: str) -> McFileTypes:
    """Tries to detect a file type"""

    if pathlib.Path(file_name).suffix == ".root":
        return McFileTypes.EIC_SMEAR

    with open(file_name, 'r') as f:
        line = ""
        try:
            i = 0
            while not line:                 # Just to fake a lot of readlines and hit the end
                line = next(f).replace('\n', '')

                # Don't go too far away if one can't determine file type
                i += 1
                if i > 20:
                    break

        except StopIteration:
            return McFileTypes.UNKNOWN

        # Determining file types
        if "BEAGLE EVENT FILE" in line:
            return McFileTypes.BEAGLE
        elif "HepMC::Version 2" in line:
            return McFileTypes.HEPMC2
        elif "HepMC::Version 3" in line:
            return McFileTypes.HEPMC3
        elif "PYTHIA EVENT FILE" in line:
            return McFileTypes.PYTHIA6_EIC
        elif len(line.split()) == 10:
            return McFileTypes.LUND_GEMC
    return McFileTypes.UNKNOWN




import logging
import time

from pyHepMC3 import HepMC3 as hm

from mcconv import detect_mc_type
from .formats.beagle import BeagleReader
from .generic_reader import GenericTextReader, UnparsedTextEvent
from .formats.lund import parse_lund_particles, parse_lund_particle_tokens
from .file_types import McFileTypes
from .eic_smear_reader import EicTreeReader, EicSmearEventData

logger = logging.getLogger("mcconv.hepmc_convert")


def hepmc_convert(input_file, output_file, input_type, hepmc_vers, progress_func=None, nskip=0, nprocess=0):

    # Choose the output format: hepmc 2 or 3 writer
    if hepmc_vers == 2 or hepmc_vers == "2" or (isinstance(hepmc_vers, str) and hepmc_vers.lower() == "hepmc2"):
        writer = hm.WriterAsciiHepMC2(output_file)
    else:
        writer = hm.WriterAscii(output_file)

    # What is the input type? Is it known?
    if not input_type or input_type == McFileTypes.UNKNOWN:
        logger.debug("Input file type is not given or UNKNOWN. Trying autodetect")

        # Input type is unknown, trying autodetection
        input_type = detect_mc_type(input_file)

    # If it is still UNKNOWN - we where unable to detect. Error raising
    if input_type == McFileTypes.UNKNOWN:
        raise ValueError("File format is UNKNOWN")

    # Set event reader according to file type
    if input_type == McFileTypes.EIC_SMEAR:
        reader = EicTreeReader()
    else:
        reader = GenericTextReader()

    # Open input file
    reader.open(input_file)

    # This is basically the same as with statement. But HepMcWriter doesn't implement __enter__() etc.
    hepmc_event = hm.GenEvent(hm.Units.GEV, hm.Units.MM)
    start_time = time.time()
    try:
        # Iterate events
        for evt_index, source_event in enumerate(reader.events(nskip, nprocess)):

            # What conversion function to use?
            if input_type == McFileTypes.EIC_SMEAR:
                eic_smear_to_hepmc(hepmc_event, source_event)
            else:
                lund_to_hepmc(hepmc_event, source_event, flavour=input_type)

            # Write event
            writer.write_event(hepmc_event)

            # call progress func, so one could work on it
            if progress_func:
                progress_func(evt_index, hepmc_event)

            hepmc_event.clear()
    finally:
        # closing everything (we are not using with statement as it is not supported by HepMC)
        writer.close()
        reader.close()
        hepmc_event.clear()
        logger.info(f"Time for the conversion = {time.time() - start_time} sec")


def add_hepmc_attribute(obj, name, str_value, val_type):
    if val_type == int:
        obj.add_attribute(name, hm.IntAttribute(int(str_value)))
    if val_type == float:
        obj.add_attribute(name, hm.FloatAttribute(float(str_value)))
    if val_type == str:
        obj.add_attribute(name, hm.StringAttribute(str_value))


def lund_to_hepmc(hepmc_evt, unparsed_event, rules=None, flavour=McFileTypes.LUND):
    """

    """
    assert isinstance(unparsed_event, UnparsedTextEvent)

    if not rules:
        prt_col_px = 6
        prt_col_py = 7
        prt_col_pz = 8
        prt_col_e = 9
        prt_col_pid = 2
        prt_col_status = 1
        evt_attrs = {"weight": (9, float)}
        prt_attrs = {}

        # Lund GEMC format has status and PID at different types
        if flavour == McFileTypes.LUND_GEMC:
            prt_col_pid = 3
            prt_col_status = 2

        if flavour == McFileTypes.PYTHIA6_EIC:
            # No event weight for pythia6
            evt_attrs = {"weight": (9, float)}

        if flavour == McFileTypes.BEAGLE:
            prt_col_px = 7
            prt_col_py = 8
            prt_col_pz = 9
            prt_col_e = 10
    else:
        prt_col_px = rules["px"]
        prt_col_py = rules["py"]
        prt_col_pz = rules["pz"]
        prt_col_e = rules["e"]
        prt_col_pid = rules["pid"]
        prt_col_status = rules["status"]
        evt_attrs = rules["evt_attrs"]
        prt_attrs = rules["prt_attrs"]

    hepmc_evt.add_attribute("start_line_index", hm.IntAttribute(unparsed_event.start_line_index))

    v1 = hm.GenVertex()
    hepmc_evt.add_vertex(v1)

    #particles = parse_lund_particles(unparsed_event)
    for particle_line in unparsed_event.unparsed_particles:

        # Parse main columns with 4 vectors
        px = float(particle_line[prt_col_px])
        py = float(particle_line[prt_col_py])
        pz = float(particle_line[prt_col_pz])
        energy = float(particle_line[prt_col_e])
        pid = int(particle_line[prt_col_pid])
        status = int(particle_line[prt_col_status])

        # Take only final state particle
        if status != 1:
            continue

        # Create a hepmc particle
        hm_particle = hm.GenParticle(hm.FourVector(px, py, pz, energy), pid, status)

        # Add particle level attributes
        for name, params in prt_attrs.items():
            column_index, field_type = params
            add_hepmc_attribute(hm_particle, name, column_index, field_type)

        # Add particle to event
        hepmc_evt.add_particle(hm_particle)

        # Add event level attributes
        for name, params in evt_attrs.items():
            column_index, field_type = params
            add_hepmc_attribute(hm_particle, name, column_index, field_type)

    return hepmc_evt


def eic_smear_to_hepmc(hepmc_evt, source_evt):
    """

    """
    assert isinstance(source_evt, EicSmearEventData)

    v1 = hm.GenVertex()
    hepmc_evt.add_vertex(v1)

    #particles = parse_lund_particles(unparsed_event)
    for particle in source_evt.particles:

        # Create a hepmc particle
        hm_particle = hm.GenParticle(hm.FourVector(particle.px, particle.py, particle.pz, particle.energy),
                                     particle.pid,
                                     particle.status)

        # Add particle to event
        hepmc_evt.add_particle(hm_particle)

    return hepmc_evt


import uproot
import os
from pprint import pprint

def _data_path(file_name):
    """Gets data file path"""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(dir_path, 'data', file_name)

local_file = uproot.open(_data_path('eic_DEMPGen_10on100_1B_1_100.root'))

#for event in local_file['EICTree']
pprint(local_file['EICTree'].typenames())

rules = {
    'id': 'event/erhic::EventMC/particles/particles.id',
    'px': 'event/erhic::EventMC/particles/particles.px',
    'py': 'event/erhic::EventMC/particles/particles.py',
    'pz': 'event/erhic::EventMC/particles/particles.pz',
    'e': 'event/erhic::EventMC/particles/particles.E',
    'ks': 'event/erhic::EventMC/particles/particles.KS',
    'vtx_x':  'event/erhic::EventMC/particles/particles.xv',
    'vtx_y':  'event/erhic::EventMC/particles/particles.yv',
    'vtx_z':  'event/erhic::EventMC/particles/particles.zv',
}

branch_names = [
    rules['id'],
    rules['px'],
    rules['py'],
    rules['pz'],
    rules['e'],
    rules['ks'],
    rules['vtx_x'],
    rules['vtx_y'],
    rules['vtx_z']
]

for events in local_file['EICTree'].iterate(branch_names, step_size=100, entry_start=0, entry_stop=2):
    for event in events:
        for particle in zip(event[rules['id']], event[rules['px']], event[rules['py']], event[rules['pz']], event[rules['e']], event[rules['ks']], event[rules['vtx_x']], event[rules['vtx_y']], event[rules['vtx_z']]):
            print(particle)
    print("-------------------------------------------")


import argparse
import glob
import logging
import sys

from .detect import detect_mc_type
from .file_types import McFileTypes, parse_file_type
from .generic_reader import UnparsedTextEvent, GenericTextReader
from .hepmc_convert import hepmc_convert


# the default ccdb logger
logger = logging.getLogger("mcconv")


def shot_progress(event_index, evt):
    if event_index and event_index % 1000 == 0:
        logger.info(f"Events processed: {event_index:<10}")

def hepmc_convert_cli():
    """Main entry point of mcconv executable"""
    parser = argparse.ArgumentParser()
    parser.add_argument('inputs', nargs='+', help="File name (wildcards allowed)")
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debugging output. More than verbose")
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose output. Info level")
    parser.add_argument("-i", "--in-type", default="", help="Input file type: [auto, pythia6-bnl]")
    parser.add_argument("-f", "--format", default="3", help="HepMC format [2,3]")
    parser.add_argument("-r", "--report", default="report", help="Report folder (set blank for no report)")
    parser.add_argument("-o", "--output", default="output.hepmc", help="File name of resulting hepmc")
    parser.add_argument("-s", "--nskip", default=0, type=int, help="Number of events to skip")
    parser.add_argument("-p", "--nprocess", default=0, type=int, help="Number of events to process")

    args = parser.parse_args()

    # Configuring logger
    # create and set console handler
    stdout_handler = logging.StreamHandler()
    stdout_handler.stream = sys.stdout
    logger.addHandler(stdout_handler)

    # create stderr handler
    stderr_handler = logging.StreamHandler()
    stderr_handler.stream = sys.stderr
    stderr_handler.setLevel(logging.ERROR)
    logger.addHandler(stderr_handler)

    # Logger level from arguments
    if args.debug:
        logger.setLevel(logging.DEBUG)
    if args.verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    # What files to process
    file_paths = []
    for user_input in args.inputs:
        # use Glob to convert something like some_dir/* to file names
        file_paths.extend([file_path for file_path in glob.glob(user_input)])

    # >oO DEBUG output
    logger.debug("Found files to process:")
    for file_path in file_paths:
        logger.debug(f"  {file_path}")

    input_file_type = parse_file_type(args.in_type)

    hepmc_convert(file_paths[0], args.output, input_file_type, args.format, shot_progress, args.nskip, args.nprocess)

if __name__ == '__main__':
    hepmc_convert_cli()



"""
Reads EicSmear tree (aka EicTree) and converts to hepmc
The class uses uproot, so it has to be installed
"""

import io
import enum
import logging

from typing import AnyStr, Callable, List
import re
import uproot
import os

# the default ccdb logger

logger = logging.getLogger("mcconv.eic_smear_reader")


class EicSmearParticleData:
    """Holds basic information on EicSmearParticle"""
    def __init__(self, pid, px, py, pz, e, status, vtx_x, vtx_y, vtx_z):
        self.pid = pid
        self.px = px
        self.py = py
        self.pz = pz
        self.energy = e
        self.status = status
        self.vtx_x = vtx_x
        self.vtx_y = vtx_y
        self.vtx_z = vtx_z


class EicSmearEventData:

    def __init__(self, particles):
        self.io_event_index = 0
        self.particles: List[EicSmearParticleData] = particles


class EicTreeReader:
    """Class EicTree reader"""

    batch_size = 1000

    def __init__(self, file=None, file_path=""):
        self.file_path = file_path
        self.file = file
        self.event_index = 0

    def open(self, file_path):
        self.file_path = file_path
        self.file = uproot.open(self.file_path)
        self.event_index = 0

    def close(self):
        if self.file:
            self.file.close()

    def events(self, evt_skip: int = 0, evt_take: int = 0):
        """Generator, reads event by event"""

        # Root files branches names:
        branch_pid = 'event/erhic::EventMC/particles/particles.id'
        branch_px = 'event/erhic::EventMC/particles/particles.px'
        branch_py = 'event/erhic::EventMC/particles/particles.py'
        branch_pz = 'event/erhic::EventMC/particles/particles.pz'
        branch_e = 'event/erhic::EventMC/particles/particles.E'
        branch_status = 'event/erhic::EventMC/particles/particles.KS'
        branch_vtx_x = 'event/erhic::EventMC/particles/particles.xv'
        branch_vtx_y = 'event/erhic::EventMC/particles/particles.yv'
        branch_vtx_z = 'event/erhic::EventMC/particles/particles.zv'

        # for convenience
        all_branches = [
            branch_pid,
            branch_px,
            branch_py,
            branch_pz,
            branch_e,
            branch_status,
            branch_vtx_x,
            branch_vtx_y,
            branch_vtx_z,
        ]

        print(self.file['EICTree'].num_entries)

        # In uproot you have to specify top index to read instead of how many events to take
        # None - take all events
        entry_stop = evt_skip + evt_take if evt_take else None

        for batch in self.file['EICTree'].iterate(all_branches, step_size=self.batch_size*10, entry_start=evt_skip, entry_stop=entry_stop, report=True):
            events, report = batch
            print(report)
            for event in events:
                particles = []
                for data in zip(event[branch_pid],
                                    event[branch_px],
                                    event[branch_py],
                                    event[branch_pz],
                                    event[branch_e],
                                    event[branch_status],
                                    event[branch_vtx_x],
                                    event[branch_vtx_y],
                                    event[branch_vtx_z]):
                    particle = EicSmearParticleData(*data)
                    particles.append(particle)
                yield EicSmearEventData(particles)
                self.event_index += 1

    def to_pandas(self, event_fields, particle_fields, chunk_size=0):
        """
        Converts events to pandas data frame.

        If
        """
        raise NotImplementedError("to_pandas function is not implemented for EicTreeReader")





import collections
from typing import List


from mcconv import GenericTextReader, UnparsedTextEvent


def _is_line_relevant(line):
    if line.strip():
        return True
    return False


def _is_event(tokens):
    return len(tokens) < 14  # Beagle event header has like 50 tokens, particle like 16


def _is_particle(tokens):
    return len(tokens) >= 14


class LundEventData:

    def __init__(self):
        """
        | 0        | Number of particles                         |
        | 1        | Mass number of the target (UD)              |
        | 2        | Atomic number oif the target (UD)           |
        | 3        | Target polarization (UD)                    |
        | 4        | Beam Polarization                           |
        | 5        | Beam type, electron=11, photon=22” (UD)     |
        | 6        | Beam energy (GeV) (UD)                      |
        | 7        | Interacted nucleon ID (2212 or 2112) (UD)   |
        | 8        | Process ID (UD)                             |
        | 9        | Event weight (UD)                           |
        """

class LundParticleData:
    """
    | 0  | line index, runs from 1 to nrTracks                                                                        |
    | 1  | status code KS (1: stable particles 11: particles which decay 55; radiative photon)                        |
    | 2  | particle KF code (211: pion, 2112:n, ....)                                                                 |
    | 3  | line number of parent particle                                                                             |
    | 4  | normally the line number of the first daughter; it is 0 for an undecayed particle or unfragmented parton   |
    | 5  | normally the line number of the last daughter; it is 0 for an undecayed particle or unfragmented parton.   |
    | 6  | px of particle                                                                                             |
    | 7  | py of particle                                                                                             |
    | 8  | pz of particle                                                                                             |
    | 9  | Energy of particle                                                                                         |
    | 10 | mass of particle                                                                                           |
    | 11 | x vertex information                                                                                       |
    | 12 | y vertex information                                                                                       |
    | 13 | z vertex information                                                                                       |
    """

    def __init__(self):
        self.id = -1
        self.status = 0
        self.pdg = 0
        self.parent_id = 0
        self.daughter_first = 0
        self.daughter_last = 0
        self.px = 0.0
        self.py = 0.0
        self.pz = 0.0
        self.energy = 0.0
        self.mass = 0.0
        self.vtx_x = 0.0
        self.vtx_y = 0.0
        self.vtx_z = 0.0


class GemcLundParticleData:

    def __init__(self):
        """
        Event record:
          column       quantity
          0         **Number of particles**
          1         Mass number of the target (UD)
          2         Atomic number oif the target (UD)
          3         Target polarization  (UD)
          4         **Beam Polarization**
          5         Beam type, electron=11, photon=22" (UD)
          6         Beam energy (GeV)  (UD)
          7         Interacted nucleon ID (2212 or 2112)  (UD)
          8         Process ID (UD)
          9         Event weight (UD)

        Particle record:
          column    quantity
          0         **index**
          1         Lifetime [nanoseconds] (UD)
          2         **status (1 is active)**
          3         **particle ID**
          4         Index of the parent (UD)
          5         Index of the first daughter (UD)
          6         **momentum x   [GeV]**
          7         **momentum y   [GeV]**
          8         **momentum z   [GeV]**
          9         Energy of the particle [GeV] (UD)
          10        Mass of the particle [GeV] (UD)
          11        **vertex x [cm]**
          12        **vertex y [cm]**
          13        **vertex z [cm]**
        """

        self.id = -1
        self.lifetime = 0
        self.status = 0
        self.pdg = 0
        self.parent_id = 0
        self.daughter_first = 0
        self.px = 0.0
        self.py = 0.0
        self.pz = 0.0
        self.energy = 0.0
        self.mass = 0.0
        self.vtx_x = 0.0
        self.vtx_y = 0.0
        self.vtx_z = 0.0


LundParticleTuple = collections.namedtuple('LundParticleTuple',
    [
        'id',
        'status',
        'pdg',
        'parent_id',
        'daughter_first',
        'daughter_last',
        'px',
        'py',
        'pz',
        'energy',
        'mass',
        'vtx_x',
        'vtx_y',
        'vtx_z',
    ])


def parse_lund_particle_tokens(particle_tokens):
    return LundParticleTuple(
    int(particle_tokens[0])     ,    # id
    int(particle_tokens[1])     ,    # status
    int(particle_tokens[2])     ,    # pdg
    int(particle_tokens[3])     ,    # parent_id
    int(particle_tokens[4])     ,    # daughter_first
    int(particle_tokens[5])     ,    # daughter_last
    float(particle_tokens[6])   ,    # px
    float(particle_tokens[7])   ,    # py
    float(particle_tokens[8])   ,    # pz
    float(particle_tokens[9])   ,    # energy
    float(particle_tokens[10])  ,    # mass
    float(particle_tokens[11])  ,    # vtx_x
    float(particle_tokens[12])  ,    # vtx_y
    float(particle_tokens[13])  ,    # vtx_z
    )



def parse_lund_particles(unparsed_event: UnparsedTextEvent):

    particles = []
    for particle_tokens in unparsed_event.unparsed_particles:
        particle = LundParticleData()
        particle.id = int(particle_tokens[0])
        particle.status = int(particle_tokens[1])
        particle.pdg = int(particle_tokens[2])
        particle.parent_id = int(particle_tokens[3])
        particle.daughter_first = int(particle_tokens[4])
        particle.daughter_last = int(particle_tokens[5])
        particle.px = float(particle_tokens[6])
        particle.py = float(particle_tokens[7])
        particle.pz = float(particle_tokens[8])
        particle.energy = float(particle_tokens[9])
        particle.mass = float(particle_tokens[10])
        particle.vtx_x = float(particle_tokens[11])
        particle.vtx_y = float(particle_tokens[12])
        particle.vtx_z = float(particle_tokens[13])
        particles.append(particle)

    return particles


def parse_gemc_lund_particles(unparsed_event: UnparsedTextEvent):

    particles = []
    for particle_line in unparsed_event.unparsed_particles:
        particle = GemcLundParticleData()
        particle.id = int(particle_line[0])
        particle.lifetime = float(particle_line[1])
        particle.status = int(particle_line[2])
        particle.pdg = int(particle_line[3])
        particle.parent_id = int(particle_line[4])
        particle.daughter_first = int(particle_line[5])
        particle.px = float(particle_line[6])
        particle.py = float(particle_line[7])
        particle.pz = float(particle_line[8])
        particle.energy = float(particle_line[9])
        particle.mass = float(particle_line[10])
        particle.vtx_x = float(particle_line[11])
        particle.vtx_y = float(particle_line[12])
        particle.vtx_z = float(particle_line[13])
        particles.append(particle)

    return particles


"""

Universal parser for text files from generators. It aims for cases:
1. Files have are space/tab separated data values
2. Each event start with one line of data
3. Then there are several lines with data for each particle in event
4. (!) number of tokens in event and particle line should be different

2-3. Means that if we see an event record after particle records
     it means, that the new event has started

 This class works for files like from pythia6, beagle, etc...

 D A T A   E X A M P L E:
 Lets look at beagle event record:
      BEAGLE EVENT FILE
      ============================================
      I, ievent, ...
      ============================================
      I  ISTHKK(I)  ...
      ============================================
      0          1    1   11    2    1    ... <50+ tokens>
      ============================================
       1     21         11        0       -0.000000       -0.000000       -5.000000  ... <15 tokens>
       2     21       2212        0        0.000000        0.000000       50.107677  ...
      ...
      20      1       2212        0        0.018354        0.019732       45.588296  ...
      =============== Event finished ===============
      0          2    1   11    2    1 ... <next event >
      ============================================
      1     21         11        0        0.000000        0.000000       -5.000000 ...
 What we see here is:
     1. First 5 lines are to be skipped
     2. Lines with '====================' are to be skipped
     3. Lines like "0 1 1 11 2 1 ..." are the beginning of event data with whole vent info
     4. All the next lines like "1 21 22122 ..." are particle data lines (called record in this class)
     5. All lines in 3 and 4 need to be tokenized
     6. In other file formats comments like '#...' should be ignored

 As a result of parsing we will have and event in terms of :
  TextFileEvent:
      started_at_line;    // Line number at which the event has started
      event_tokens;       // All tokens like ["0", "1", "1", "11", "2", "1", ...] from 'event' related line/s
      record_tokens       // All tokens like ["1", "21", "22122", ...] from each particle or 'record' lines


"""
import io
import enum
import shlex
from typing import AnyStr, Callable
import re


class UnparsedTextEvent:
    """Unparsed """
    def __init__(self):
        self.start_line_index = 0
        self.io_event_index = 0             # event number as it was read from file
        self.event_tokens = []
        self.unparsed_particles = []

FIELDS_PATTERN = re.compile(r"(?:\"(.*?)\"|(\S+))")

def regex_split(data):
    return [x[0] or x[1] for x in FIELDS_PATTERN.findall(data)]

class GenericTextReader:
    """Generic class to read EIC related text MC file formats"""

    buffer_size = 65536

    def __init__(self, is_line_relevant: Callable = None, is_event: Callable = None, is_particle: Callable = None, convert_rules={}, skip_lines: int = 0):
        self.file_path = ""
        self.file: [io.TextIOBase, None] = None
        self.line_index = 0
        self.skip_lines = skip_lines
        self.is_line_relevant = is_line_relevant
        self.is_event_line = is_event
        self.is_particle_line = is_particle
        self.last_record_was_particle = False
        self.particle_tokens_len = 14
        self.hepmc_conv_rules = convert_rules
        self.event_index = 0

        if not is_line_relevant:
            self.is_line_relevant = self.default_is_line_relevant

        if not is_event:
            self.is_event_line = self.default_is_event

        if not is_particle:
            self.is_particle_line = self.default_is_particle

    def open(self, file_path):
        self.file_path = file_path
        self.file = open(self.file_path)

    def close(self):
        if self.file:
            self.file.close()

    def default_is_line_relevant(self, line):
        is_comment = any(word in line for word in ['BEAGLE', 'I', '='])
        return not is_comment

    def default_is_event(self, tokens):
        return len(tokens) != self.particle_tokens_len  # Beagle event header has like 50 tokens, particle like 16

    def default_is_particle(self, tokens):
        return len(tokens) == self.particle_tokens_len

    def events(self, evt_skip: int = 0, evt_take: int = 0):
        """Generator, reads event by event"""

        event: UnparsedTextEvent = UnparsedTextEvent()

        # Number of events processed
        self.event_index = 0

        while True:
            lines = self.file.readlines(self.buffer_size)
            if not lines:
                break
            for line in lines:
                assert isinstance(line, str)
                line = line.strip()

                # Checks if the line is not empty, is comment, or in skip range
                if self._should_skip_line(line):
                    self.line_index += 1
                    continue

                # tokenize line
                #tokens = regex_split(line)
                tokens = line.replace('\t', ' ').split()

                # Is it a beginning of a new event?
                if self.is_event_line(tokens):

                    # It is not the first event, right?
                    if self.last_record_was_particle:

                        try:
                            # Should we skip the event?
                            if evt_skip <= self.event_index:

                                # Select the event
                                yield event

                                # Should we break?
                                if evt_take and evt_take + evt_skip == self.event_index + 1:
                                    return
                        finally:
                            # Whatever we decide
                            # Create a new event class
                            event = UnparsedTextEvent()
                            self.event_index += 1

                    # Add tokens
                    event.start_line_index = self.line_index
                    event.io_event_index = self.event_index
                    event.event_tokens = tokens
                    self.last_record_was_particle = False

                # It is a particle record
                if self.is_particle_line(tokens):
                    self.last_record_was_particle = True
                    event.unparsed_particles.append(tokens)

                # increment line count in the end
                self.line_index += 1

        # End... Do we have some event prepared?
        if event.event_tokens or event.unparsed_particles:
            yield event

        return


    def to_pandas(self, event_fields, particle_fields, chunk_size=0):
        """
        Converts events to pandas data frame.

        If
        """


    def unparsed_events(self):
        """
         Same as events, but events are being overwritten by child classes

        :return: Generator. List(UnparsedTextEvent)
        """

    @property
    def reader_name(self):
        """
        The name of the reader class
        :return: A string with the reader name
        """
        return "generic_reader"

    def _should_skip_line(self, line):
        """Checks if the line is not empty, is comment, or in skip range"""
        if not line:
            return True

        if self.line_index < self.skip_lines:
            return True

        return not self.is_line_relevant(line)



from mcconv import GenericTextReader, UnparsedTextEvent


def _is_line_relevant(line):
    is_comment = any(word in line for word in ['BEAGLE', 'I', '='])
    return not is_comment


def _is_event(tokens):
    return len(tokens) > 20  # Beagle event header has like 50 tokens, particle like 16


def _is_particle(tokens):
    return len(tokens) < 20


class BeagleReader(GenericTextReader):

    def __init__(self):
        super().__init__(_is_line_relevant, _is_event, _is_particle)

    def unparsed_events(self):
        """Just a proxy for an unparsed events"""
        for event in super().events():
            yield event

    def events(self):
        """Generator, reads event by event"""

        for event in super().events():
            assert isinstance(event, UnparsedTextEvent)





class BeagleEventData:

    def __init__(self, unparsed_event):

        self.unparsed_event = unparsed_event

        self.event_index = -1              # Usually it what people call 'event number' but the 'number' is very misleading
        self.started_at_line = -1

        self.genevent: int = 0             # 2 I  trials to generate this event
        self.ltype: int = 0                # 3 I  particle code for incoming lepton
        self.a_targ: int = 0               # 4 I  A of the target nucleus
        self.z_targ: int = 0               # 5 I  Z of the target nucleus
        self.pzlep: float = 0              # 6 D  lab pz of the lepton beam (sign(pz)*p for non-zero crossing angle)
        self.pztarg: float = 0             # 7 D  lab pz/A of the ion beam (sign(pz)*p/A for non-zero crossing angle)
        self.pznucl: float = 0             # 8 D  lab pz of the struck nucleon (sign(pz)*p for non-zero crossing angle)
        self.crang: float = 0              # 9 D  crossing-angle (mr). crang = (1000*) atan(px/pz) for the beam momentum with nonzero px. Note: we assume one of the beams has px=py=0 and the other py=0.
        self.crori: int = 0                # 10 I  crossing angle orientation. lepton beam defines \pm\hat{z} direction. \pm 2 hadron/ion beam defines
                                           # \pm\hat{z} direction. 0 means no crossing angle.
        self.subprocess: int = 0           # 11 I   pythia subprocess (MSTI(1)), for details see PYTHIA
        self.nucleon: int = 0              # 12 I   hadron beam type (MSTI(12))
        self.targetparton: int = 0         # 13 I  parton hit in the target (MSTI(16))
        self.xtargparton: float = 0        # 14 D  x of target parton (PARI(34))
        self.beamparton: int = 0           # 15 I  in case of resolved photon processes and soft VMD the virtual photon has a hadronic structure. This gives the info which parton interacted with the target parton (MSTI(15))
        self.xbeamparton: float = 0        # 16 D  x of beam parton (PARI(33))
        self.thetabeamparton: float = 0    # 17 D  theta of beam parton (PARI(53))
        self.truey: float = 0              # 18 D  are the kinematic variables of the event.
        self.trueQ2: float = 0             # 19 D  are the kinematic variables of the event.
        self.truex: float = 0              # 20 D  are the kinematic variables of the event.
        self.trueW2: float = 0             # 21 D  are the kinematic variables of the event.
        self.trueNu: float = 0             # 22 D  are the kinematic variables of the event.
                                           #       If radiative corrections are turned on they are different from what is calculated from the scattered lepton.
                                           #       If radiative corrections are turned off they are the same as what is calculated from the scattered lepton
        self.leptonphi: float = 0          # 23 D  phi of the lepton (VINT(313)) in the laboratory frame
        self.s_hat: float = 0              # 24 D  s-hat of the process (PARI(14))
        self.t_hat: float = 0              # 25 D  Mandelstam t (PARI(15))
        self.u_hat: float = 0              # 26 D  Mandelstam u (PARI(16))
        self.pt2_hat: float = 0            # 27 D  pT2-hat of the hard scattering (PARI(18))
        self.Q2_hat: float = 0             # 28 D  Q2-hat of the hard scattering (PARI(22)),
        self.F2: float = 0                 # 29 D  information used and needed in the radiative correction code
        self.F1: float = 0                 # 30 D  information used and needed in the radiative correction code
        self.R: float = 0                  # 31 D  information used and needed in the radiative correction code
        self.sigma_rad: float = 0          # 32 D  information used and needed in the radiative correction code
        self.SigRadCor: float = 0          # 33 D  information used and needed in the radiative correction code
        self.EBrems: float = 0             # 34 D  energy of the radiative photon in the nuclear rest frame
        self.photonflux: float = 0         # 35 D  flux factor from PYTHIA (VINT(319))
        self.b: float = 0                  # 36 D  Impact parameter (in fm), radial position of virtual photon w.r.t. the center of the nucleus in the nuclear TRF with z along γ*
        self.Phib: float = 0               # 37 D  azimuthal angle of the impact parameter of the virtual photon in the nuclear TRF with z along γ* and φ=0 for the scattered electron
        self.Thickness: float = 0          # 38 D  Nuclear thickness T(b)/ρ0 in units of fm.
        self.ThickScl: float = 0           # 39 D  Nuclear thickness T(b) in nucleons/fm2.
        self.Ncollt: int = 0               # 40 I  Number of collisions between the incoming γ* and nucleons in the nucleus (same as number of participating nucleons)
        self.Ncolli: int = 0               # 41 I  Number of inelastic collisions between the incoming γ* and nucleons in the nucleus (same as number of inelastically participating nucleons)
        self.Nwound: int = 0               # 42 I  Number of wounded nucleons, including those involved in the intranuclear cascade. WARNING: This variable is not finalized yet
        self.Nwdch: int = 0                # 43 I  Number of wounded protons, including those involved in the intranuclear cascade. WARNING: This variable is not finalized yet
        self.Nnevap: int = 0               # 44 I  Number of evaporation (and nuclear breakup) neutrons
        self.Npevap: int = 0               # 45 I  Number of evaporation (and nuclear breakup) protons
        self.Aremn: int = 0                # 46 I  A of the nuclear remnant after evaporation and breakup
        self.Ninc: int = 0                 # 47 I  Number of stable hadrons from the Intra Nuclear Cascade
        self.Nincch: int = 0               # 48 I  Number of charged stable hadrons from the Intra Nuclear Cascade
        self.d1st: float = 0               # 49 D  density-weighted distance from first collision to the edge of the nucleus (amount of material traversed / ρ0)
        self.davg: float = 0               # 50 D  Average density-weighted distance from all inelastic collisions to the edge of the nucleus
        self.pxf: float = 0                # 51 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
        self.pyf: float = 0                # 52 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
        self.pzf: float = 0                # 53 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
        self.Eexc: float = 0               # 54 D  Excitation energy in the nuclear remnant before evaporation and breakup.
        self.RA: float = 0                 # 55 D  Nuclear PDF ratio for the up sea for the given event kinematics (x,Q2), but set to 1 if multi-nucleon shadowing is off (genShd=1)
        self.User1: float = 0              # 56 D  User variables to prevent/delay future format changes
        self.User2: float = 0              # 57 D  User variables to prevent/delay future format changes
        self.User3: float = 0              # 58 D  User variables to prevent/delay future format changes
        self.nrTracks: int = 0             # 59 I  number of tracks in this event, including event history


def beagle_parse_event(unparsed_event: UnparsedTextEvent) -> BeagleEventData:
    event_tokens = unparsed_event.event_tokens
    event = BeagleEventData(unparsed_event)

    # event_tokens[0] I  line num (always zero in file)
    event.event_index  = int(event_tokens[1])         # 1  I ievent - event num
    event.genevent     = int(event_tokens[2])         # 2  I trials to generate this event
    event.ltype        = int(event_tokens[3])         # 3  I particle code for incoming lepton
    event.a_targ       = int(event_tokens[4])         # 4  I A of the target nucleus
    event.z_targ       = int(event_tokens[5])         # 5  I Z of the target nucleus
    event.pzlep        = float(event_tokens[6])       # 6  D lab pz of the lepton beam (sign(pz)*p for non-zero crossing angle)
    event.pztarg       = float(event_tokens[7])       # 7  D lab pz/A of the ion beam (sign(pz)*p/A for non-zero crossing angle)
    event.pznucl       = float(event_tokens[8])       # 8  D lab pz of the struck nucleon (sign(pz)*p for non-zero crossing angle)
    event.crang        = float(event_tokens[9])       # 9  D crossing-angle (mr). crang = (1000*) atan(px/pz) for the beam momentum with nonzero px. Note: we assume one of the beams has px=py=0 and the other py=0.
    event.crori        = int(event_tokens[10])        # 10 I  crossing angle orientation. lepton beam defines \pm\hat{z} direction. \pm 2 hadron/ion beam defines                                                                    #    \pm\hat{z} direction. 0 means no crossing angle.
    event.subprocess   = int(event_tokens[11])        # 11 I   pythia subprocess (MSTI(1)), for details see PYTHIA
    event.nucleon      = int(event_tokens[12])        # 12 I   hadron beam type (MSTI(12))
    event.targetparton =    int(event_tokens[13])     # 13 I  parton hit in the target (MSTI(16))
    event.xtargparton =     float(event_tokens[14])   # 14 D  x of target parton (PARI(34))
    event.beamparton =      int(event_tokens[15])     # 15 I  in case of resolved photon processes and soft VMD the virtual photon has a hadronic structure. This gives the info which parton interacted with the target parton (MSTI(15))
    event.xbeamparton =     float(event_tokens[16])   # 16 D  x of beam parton (PARI(33))
    event.thetabeamparton = float(event_tokens[17])   # 17 D  theta of beam parton (PARI(53))
    event.truey =           float(event_tokens[18])   # 18 D  are the kinematic variables of the event.
    event.trueQ2 =          float(event_tokens[19])   # 19 D  are the kinematic variables of the event.
    event.truex =           float(event_tokens[20])   # 20 D  are the kinematic variables of the event.
    event.trueW2 =          float(event_tokens[21])   # 21 D  are the kinematic variables of the event.
    event.trueNu =          float(event_tokens[22])   # 22 D  are the kinematic variables of the event.
    event.leptonphi =       float(event_tokens[23])   # 23 D  phi of the lepton (VINT(313)) in the laboratory frame
    event.s_hat =           float(event_tokens[24])   # 24 D  s-hat of the process (PARI(14))
    event.t_hat =           float(event_tokens[25])   # 25 D  Mandelstam t (PARI(15))
    event.u_hat =           float(event_tokens[26])   # 26 D  Mandelstam u (PARI(16))
    event.pt2_hat =         float(event_tokens[27])   # 27 D  pT2-hat of the hard scattering (PARI(18))
    event.Q2_hat =          float(event_tokens[28])   # 28 D  Q2-hat of the hard scattering (PARI(22)),
    event.F2 =              float(event_tokens[29])   # 29 D  information used and needed in the radiative correction code
    event.F1 =              float(event_tokens[30])   # 30 D  information used and needed in the radiative correction code
    event.R =               float(event_tokens[31])   # 31 D  information used and needed in the radiative correction code
    event.sigma_rad =       float(event_tokens[32])   # 32 D  information used and needed in the radiative correction code
    event.SigRadCor =       float(event_tokens[33])   # 33 D  information used and needed in the radiative correction code
    event.EBrems =          float(event_tokens[34])   # 34 D  energy of the radiative photon in the nuclear rest frame
    event.photonflux =      float(event_tokens[35])   # 35 D  flux factor from PYTHIA (VINT(319))
    event.b =               float(event_tokens[36])   # 36 D  Impact parameter (in fm), radial position of virtual photon w.r.t. the center of the nucleus in the nuclear TRF with z along γ*
    event.Phib =            float(event_tokens[37])   # 37 D  azimuthal angle of the impact parameter of the virtual photon in the nuclear TRF with z along γ* and φ=0 for the scattered electron
    event.Thickness =       float(event_tokens[38])   # 38 D  Nuclear thickness T(b)/ρ0 in units of fm.
    event.ThickScl =        float(event_tokens[39])   # 39 D  Nuclear thickness T(b) in nucleons/fm2.
    event.Ncollt =          int(event_tokens[40])     # 40 I  Number of collisions between the incoming γ* and nucleons in the nucleus (same as number of participating nucleons)
    event.Ncolli =          int(event_tokens[41])     # 41 I  Number of inelastic collisions between the incoming γ* and nucleons in the nucleus (same as number of inelastically participating nucleons)
    event.Nwound =          int(event_tokens[42])     # 42 I  Number of wounded nucleons, including those involved in the intranuclear cascade. WARNING: This variable is not finalized yet
    event.Nwdch =           int(event_tokens[43])     # 43 I  Number of wounded protons, including those involved in the intranuclear cascade. WARNING: This variable is not finalized yet
    event.Nnevap =          int(event_tokens[44])     # 44 I  Number of evaporation (and nuclear breakup) neutrons
    event.Npevap =          int(event_tokens[45])     # 45 I  Number of evaporation (and nuclear breakup) protons
    event.Aremn =           int(event_tokens[46])     # 46 I  A of the nuclear remnant after evaporation and breakup
    event.Ninc =            int(event_tokens[47])     # 47 I  Number of stable hadrons from the Intra Nuclear Cascade
    event.Nincch =          int(event_tokens[48])     # 48 I  Number of charged stable hadrons from the Intra Nuclear Cascade
    event.d1st =            float(event_tokens[49])   # 49 D  density-weighted distance from first collision to the edge of the nucleus (amount of material traversed / ρ0)
    event.davg =            float(event_tokens[50])   # 50 D  Average density-weighted distance from all inelastic collisions to the edge of the nucleus
    event.pxf =             float(event_tokens[51])   # 51 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
    event.pyf =             float(event_tokens[52])   # 52 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
    event.pzf =             float(event_tokens[53])   # 53 D  Fermi-momentum of the struck nucleon (or sum Fermi momentum for all inelastic nucleon participants for Ncolli>1) in target rest frame with z along gamma* direction
    event.Eexc =            float(event_tokens[54])   # 54 D  Excitation energy in the nuclear remnant before evaporation and breakup.
    event.RA =              float(event_tokens[55])   # 55 D  Nuclear PDF ratio for the up sea for the given event kinematics (x,Q2), but set to 1 if multi-nucleon shadowing is off (genShd=1)
    event.User1 =           float(event_tokens[56])   # 56 D  User variables to prevent/delay future format changes
    event.User2 =           float(event_tokens[57])   # 57 D  User variables to prevent/delay future format changes
    event.User3 =           float(event_tokens[58])   # 58 D  User variables to prevent/delay future format changes
    event.nrTracks =        int(event_tokens[59])     # 59 I  number of tracks in this event, including event history
    return event


# Translated into python from testPolarization.cc
# garren@fnal.gov, Oct. 2010
# andrii.verbytskyi@mpp.mpg.gov, Nov. 2018
#
# In this example we will place the following event into HepMC "by hand"
#
#     name status pdg_id  parent Px       Py    Pz       Energy      Mass
#  1  !p+!    3   2212    0,0    0.000    0.000 7000.000 7000.000    0.938
#  2  !p+!    3   2212    0,0    0.000    0.000-7000.000 7000.000    0.938
# =========================================================================
#  3  !d!     3      1    1,1    0.750   -1.569   32.191   32.238    0.000
#  4  !u~!    3     -2    2,2   -3.047  -19.000  -54.629   57.920    0.000
#  5  !W-!    3    -24    1,2    1.517   -20.68  -20.605   85.925   80.799
#  6  !gamma! 1     22    1,2   -3.813    0.113   -1.833    4.233    0.000
#  7  !d!     1      1    5,5   -2.445   28.816    6.082   29.552    0.010
#  8  !u~!    1     -2    5,5    3.962  -49.498  -26.687   56.373    0.006


import sys

#sys.path = update_path()

import random, math
#from pyHepMC3TestUtils import COMPARE_ASCII_FILES
from pyHepMC3 import HepMC3 as hm


def test_Polarization():
    xout1 = hm.WriterAscii("testPolarization1.dat")
    xout2 = hm.WriterAscii("testPolarization2.dat")
    xout4 = hm.WriterAsciiHepMC2("testPolarization4.out")
    xout5 = hm.WriterAscii("testPolarization5.out")

    # build the graph, which will look like
    #                       p7                   #
    # p1                   /                     #
    #   \v1__p3      p5---v4                     #
    #         \_v3_/       \                     #
    #         /    \        p8                   #
    #    v2__p4     \                            #
    #   /            p6                          #
    # p2                                         #
    #
    # define a flow pattern as  p1 . p3 . p6
    #                       and p2 . p4 . p5
    #

    # First create the event container, with Signal Process 20, event number 1
    #
    evt = hm.GenEvent(hm.Units.GEV, hm.Units.MM)
    evt.set_event_number(1)
    evt.add_attribute("signal_process_id", hm.IntAttribute(20))
    v1 = hm.GenVertex()
    evt.add_vertex(v1)
    p1 = hm.GenParticle(hm.FourVector(0, 0, 7000, 7000), 2212, 3)
    evt.add_particle(p1)
    p1.add_attribute("flow1", hm.IntAttribute(231))
    p1.add_attribute("flow1", hm.IntAttribute(231))
    p1.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p1.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))

    v2 = hm.GenVertex()
    evt.add_vertex(v2)
    p2 = hm.GenParticle(hm.FourVector(0, 0, -7000, 7000), 2212, 3)
    evt.add_particle(p2)
    p2.add_attribute("flow1", hm.IntAttribute(243))
    p2.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p2.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))
    v2.add_particle_in(p2)

    p3 = hm.GenParticle(hm.FourVector(0.750, -1.569, 32.191, 32.238), 1, 3)
    evt.add_particle(p3)
    p3.add_attribute("flow1", hm.IntAttribute(231))
    p3.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p3.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))
    v1.add_particle_out(p3)
    p4 = hm.GenParticle(hm.FourVector(-3.047, -19.0, -54.629, 57.920), -2, 3)
    evt.add_particle(p4)
    p4.add_attribute("flow1", hm.IntAttribute(243))
    p4.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p4.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))
    v2.add_particle_out(p4)

    v3 = hm.GenVertex()
    evt.add_vertex(v3)
    v3.add_particle_in(p3)
    v3.add_particle_in(p4)
    p6 = hm.GenParticle(hm.FourVector(-3.813, 0.113, -1.833, 4.233), 22, 1)
    evt.add_particle(p6)
    p6.add_attribute("flow1", hm.IntAttribute(231))
    p6.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p6.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))
    v3.add_particle_out(p6)
    p5 = hm.GenParticle(hm.FourVector(1.517, -20.68, -20.605, 85.925), -24, 3)
    evt.add_particle(p5)
    p5.add_attribute("flow1", hm.IntAttribute(243))
    p5.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p5.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi * 2))
    v3.add_particle_out(p5)
    # create v4
    v4 = hm.GenVertex(hm.FourVector(0.12, -0.3, 0.05, 0.004))
    evt.add_vertex(v4)
    v4.add_particle_in(p5)
    p7 = hm.GenParticle(hm.FourVector(-2.445, 28.816, 6.082, 29.552), 1, 1)
    evt.add_particle(p7)
    v4.add_particle_out(p7)
    p8 = hm.GenParticle(hm.FourVector(3.962, -49.498, -26.687, 56.373), -2, 1)
    evt.add_particle(p8)
    v4.add_particle_out(p8)

    evt.add_attribute("signal_process_vertex", hm.IntAttribute(v3.id()))
    # the event is complete, we now print it out
    hm.Print.content(evt)
    hm.Print.listing(evt, 8)
    print(hm.version())
    print(hm.Print.line(v4, True))
    a = 0
    print(evt.particles())
    print(len(evt.particles()))
    print(evt.particles()[0])
    for ip in evt.particles():
        print(hm.Print.line(ip, True))
    xout1.write_event(evt)

    # write event in old format
    xout4.write_event(evt)
    # make a copy and write it
    xout5.write_event(hm.GenEvent(evt))
    # try changing polarization
    p2.add_attribute("theta", hm.DoubleAttribute(random.random() * math.pi))
    p2.add_attribute("phi", hm.DoubleAttribute(random.random() * math.pi))
    xout2.write_event(evt)
    xout1.close()
    xout2.close()
    xout4.close()
    xout5.close()
    # now clean-up by deleteing all objects from memory
    #
    # deleting the event deletes all contained vertices, and all particles
    # contained in those vertices
    evt.clear()

if __name__ == "__main__":
    result = 1
    test_Polarization()
    sys.exit(result)



import enum


class McFileTypes (enum.Enum):
    """Enumeration of known formats (unique strings)

    (!) Warning (!) Inside this library use 'import .file_types' not 'import mcconv.file_types'
    https://stackoverflow.com/a/61395054/548894

    Another thing on enums - values are uppercase because they are constants
    https://stackoverflow.com/a/21359677/548894
    """
    UNKNOWN = "unknown"
    BEAGLE = "beagle"
    HEPMC2 = "hepmc2"
    HEPMC3 = "hepmc3"
    LUND = "pythia_lund"
    LUND_GEMC = "lund_gemc"
    PYTHIA6_EIC = "pythia_bnl"
    EIC_SMEAR = "eic_smear"




def parse_file_type(in_string, raise_on_unknown=True):
    if not in_string or in_string == "unknown":
        return McFileTypes.UNKNOWN
    if in_string == "beagle":
        return McFileTypes.BEAGLE
    if in_string == "hepmc2":
        return McFileTypes.HEPMC2
    if in_string == "hepmc3":
        return McFileTypes.HEPMC300
    if in_string == "pythia_bnl":
        return McFileTypes.PYTHIA6_EIC
    if in_string == "pythia_lund":
        return McFileTypes.LUND
    if in_string == "pythia_gemc":
        return McFileTypes.LUND_GEMC

    if raise_on_unknown:
        message = f"File type='{in_string}' is unknown"
        raise ValueError(message)

    return McFileTypes.UNKNOWN


from mcconv import detect_mc_type

import os
import unittest
from mcconv import detect_mc_type, McFileTypes


class TestStringMethods(unittest.TestCase):

    def setUp(self) -> None:
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.data_path = os.path.join(dir_path, 'data')

    def _get_file_path(self, path):
        """Gets data file path"""
        return os.path.join(self.data_path, path)


    def test_beagle(self):
        """Test detecting BeaGLE file type"""

        result = detect_mc_type(self._get_file_path('beagle_3events.txt'))
        self.assertEqual(result, McFileTypes.BEAGLE)

    def test_hepmc2(self):
        """Test detecting HepMC2 file type"""

        result = detect_mc_type(self._get_file_path('hepmc2.hepmc'))
        self.assertEqual(result, McFileTypes.HEPMC2)

    def test_hepmc3(self):
        """Test detecting BeaGLE file type"""

        result = detect_mc_type(self._get_file_path('hepmc3.hepmc'))
        self.assertEqual(result, McFileTypes.HEPMC3)

    def test_lund(self):
        result = detect_mc_type(self._get_file_path('gemc-lund.3evt.txt'))
        self.assertEqual(result, McFileTypes.LUND_GEMC)

    def test_pythia6radcor(self):
        result = detect_mc_type(self._get_file_path('pythia6-radcor-10evt.txt'))
        self.assertEqual(result, McFileTypes.PYTHIA6_EIC)

    def test_eic_smear(self):
        result = detect_mc_type(self._get_file_path('eic_DEMPGen_10on100_1B_1_100.root'))
        self.assertEqual(result, McFileTypes.EIC_SMEAR)

if __name__ == '__main__':
    unittest.main()


from mcconv import GenericTextReader, UnparsedTextEvent, hepmc_convert

import os
import unittest
from mcconv import detect_mc_type, McFileTypes


class TestGenericTextReader(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def _data_path(self, file_name):
        """Gets data file path"""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dir_path, 'data', file_name)

    def _tmp_path(self, file_name):
        """Gets output temporary file path"""

        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'tmp', file_name)

    def test_convert_file(self):
        """Test detecting BeaGLE file type"""

        hepmc_convert(self._data_path('gemc-lund.3evt.txt'),
                      self._tmp_path('lund-convert.hepmc'),
                      McFileTypes.LUND_GEMC, "hepmc3")
